import RenderInfo from "./util/RenderInfo.js";
import Transform from "./Transform.js";

export default class GameObject {
    constructor(name=null) {
        this.name = name;
        this.layer = 0;
        this.transform = new Transform(this);
        this.components = [];
        this.active = true;
        this.destroyed = false;
    }

    toString() {
        return `${this.name || "GameObject"}(at ${this.transform.position.toString()})`;
    }

    /**
     * Initialize this GameObject, all its components and the all its children recursively
     * 
     * @param {*} gl The WebGL context
     */
    init(gl) {
        this.components.forEach((comp) => {
            if (comp.init)
                comp.init(gl);
        });
        this.transform.children.forEach((child) => {
            if (child.gameObject.init)
                child.gameObject.init(gl);
        });
    }

    /**
     * Update the GameObject, all its components and children recursively
     * 
     * @param {Number} deltaTime The delta time since the last call to update
     */
    update(deltaTime) {
        if (!this.active)
            return;
        
        this.components.forEach((comp) => {
            if (comp.update)
                comp.update(deltaTime);
        });
        this.transform.children.forEach((child) => {
            if (child.gameObject.update)
                child.gameObject.update(deltaTime);
        });
    }

    /**
     * Render the game object, all its components and children recursively
     * 
     * @param {RenderInfo} renderInfo Render info context
     */
    render(renderInfo) {
        if (!this.active)
            return;
        
        // save the current model view matrix state and apply own transform
        renderInfo.modelViewStack.save();
        renderInfo.modelViewStack.multiply(this.transform.glMat4);

        this.components.forEach((comp) => {
            if (comp.render) {
                comp.render(renderInfo);
            }
        });

        this.transform.children.forEach((child) => {
            if (child.gameObject.render) {
                child.gameObject.render(renderInfo);
            }
        });
        
        renderInfo.modelViewStack.restore();
    }

    destroy() {
        this.destroyed = true;
    }

    addComponent(component) {
        this.components.push(component);
        component.gameObject = this;
    }

    removeComponent(component) {
        const idx = this.components.indexOf(component);

        if (idx !== -1) {
            const component = this.components.splice(idx);
            component.gameObject = null;
        }
    }

    getComponent(type) {
        const c = this.components.filter((component) => component instanceof type);
        if (c.length > 0)
            return c[0];
        return null;
    }

    hasComponent(type) {
        return this.getComponent(type) != null;
    }

    setActive(active) {
        this.active = active;
        console.log(`${this.name} setActive(${active})`);
    }
}