import Vector3 from "./math/Vector3.js";
import Matrix4 from "./math/Matrix4.js";
import Quaternion from "./math/Quaternion.js";

export default class Transform {
    constructor(gameObject) {
        this.gameObject = gameObject;
        this.parent = null;
        this.position = new Vector3();
        this.rotation = new Quaternion();
        this.scale = Vector3.one;
        this.children = [];
        this.mat4 = new Matrix4();
        this.updateMat4();
    }

    /**
     * Build an affine mat4 transform that represents the current state
     */
    updateMat4() {
        this.mat4 = Matrix4.fromRotationTranslationScale(this.rotation, this.position, this.scale);
    }

    get glMat4() {
        return this.mat4.glMat4;
    }

    get rotationMat4() {
        return Matrix4.fromQuat(this.rotation);
    }

    get worldMat4() {
        return this.worldMat4.glMat4;
    }

    get worldMat4() {
        return this.parent ? this.parent.worldMat4.multiply(this.mat4) : this.mat4;
    }

    get worldPosition() {
        return this.worldMat4.getTranslation();
    }

    get worldRotation() {
        return this.worldMat4.getRotation();
    }

    get worldRotationMat4() {
        return Matrix4.fromQuat(this.worldRotation);
    }

    addChild(child) {
        this.children.push(child);
        child.parent = this;
    }

    removeChild(child) {
        const idx = this.children.indexOf(child);

        if (idx !== -1) {
            const child = this.children.splice(idx);
            child.parent = null;
        }
    }

    setPosition(x, y, z) {
        this.position = new Vector3(x, y, z);
        this.updateMat4();
    }

    getRotation() {
        return this.rotation.clone();
    }

    setRotation(roll, pitch, yaw) {
        this.rotation = Quaternion.fromEuler(roll, pitch, yaw);
        this.updateMat4();
    }

    setRotationFromQuat(quat) {
        this.rotation = quat.clone();
        this.updateMat4();
    }

    setScale(x, y, z) {
        this.scale = new Vector3(x, y, z);
        this.updateMat4();
    }

    translate(x, y, z) {
        this.position = this.position.add(new Vector3(x, y, z));
        this.updateMat4();
    }

    get forward() {
        return this.worldRotationMat4.multiply(Vector3.forward);
    }

    get up() {
        return this.worldRotationMat4.multiply(Vector3.up);
    }

    get right() {
        return this.worldRotationMat4.multiply(Vector3.right);
    }
}