import {getPoseMatrix} from './util/MatrixUtil.js';
import MatrixStack from './util/MatrixStack.js';
import RenderInfo from './util/RenderInfo.js';
import Camera from './Camera.js';
import Debug from './Debug.js';
import Input from './Input.js';

// Vertex shader program
const VS_SOURCE = `
    attribute vec4 aVertexPosition;
    attribute vec4 aVertexColor; 
    uniform mat4 uModelViewMatrix;
    uniform mat4 uProjectionMatrix;
    varying lowp vec4 vColor;
    void main(void) {
        gl_Position =
            uProjectionMatrix * uModelViewMatrix * aVertexPosition;
        vColor = aVertexColor;
    }
`;
// Fragment shader program
const FS_SOURCE = `
    varying lowp vec4 vColor;
    void main(void) {
        gl_FragColor = vColor;
    }
`;

export class VRoom {
    constructor() {
        this.gl = null;
        this.programInfo = null;
        this.vrDisplay = null;
        this.canvas = null;
        this.canvasId = null;
        this.currentScene = null;
        this.lastUpdate = 0;
        this.eventListeners = {};
    }

    addEventListener(name, listener) {
        this.eventListeners[name] = [...(this.eventListeners[name] || []), listener];
    }

    removeEventListener(name, listener) {
        this.eventListeners[name] = (this.eventListeners[name] || []).filter((l) => l != listener);
    }

    dispatchEvent(name) {
        if (this.eventListeners[name]) {
            this.eventListeners[name].forEach((listener) => listener(this));
        } else {
            console.warn(`No event listener for '${name}' found.`);
        }
    }

    start() {
        console.log("start");
        // Wait until we have a WebGL context to resize and start rendering.
        window.addEventListener("resize", () => this.onResize, false);
        this.onResize();
    }

    init() {
        this.canvas = document.querySelector(this.canvasId);
        Input.initCanvasInput(this.canvas);
        this.initVR().then(() => {
            Debug.init(this.gl);
            if (this.currentScene) {
                this.currentScene.init(this.gl);
            }
        })
    }

    initGL(preserve_drawing_buffer) {
        /*
        Initialize a shader program, so WebGL knows how to draw our data
        */
        function initShaderProgram(gl, vsSource, fsSource) {
            const vertexShader = loadShader(gl, gl.VERTEX_SHADER, vsSource);
            const fragmentShader = loadShader(gl, gl.FRAGMENT_SHADER, fsSource);
            // Create the shader program
            const shaderProgram = gl.createProgram();
            gl.attachShader(shaderProgram, vertexShader);
            gl.attachShader(shaderProgram, fragmentShader);
            gl.linkProgram(shaderProgram);
            // If creating the shader program failed, alert
            if (!gl.getProgramParameter(shaderProgram, gl.LINK_STATUS)) {
                alert('Unable to initialize the shader program: ' +
                    gl.getProgramInfoLog(shaderProgram));
                return null;
            }
            return shaderProgram;
        }
        /*
        creates a shader of the given type, uploads the source and
        compiles it.
        */
        function loadShader(gl, type, source) {
            const shader = gl.createShader(type);
            // Send the source to the shader object
            gl.shaderSource(shader, source);
            // Compile the shader program
            gl.compileShader(shader);
            // See if it compiled successfully
            if (!gl.getShaderParameter(shader, gl.COMPILE_STATUS)) {
                alert('An error occurred compiling the shaders: ' +
                    gl.getShaderInfoLog(shader));
                gl.deleteShader(shader);
                return null;
            }
            return shader;
        }

        var gl_attribs = {
            alpha: false,
            antialias: false,
            preserveDrawingBuffer: preserve_drawing_buffer
        };
        this.gl = this.canvas.getContext("webgl", gl_attribs);
        this.canvas.getContext("webgl", gl_attribs) ||
        this.canvas.getContext("experimental-webgl", gl_attribs);
        // If we don't have a GL context, give up now
        if (!this.gl) {
            alert(
                "Unable to initialize WebGL. Your browser or machine " +
                "may not support it."
            );
            return;
        }
        const shaderProgram = initShaderProgram(this.gl, VS_SOURCE, FS_SOURCE);
        this.programInfo = {
            program: shaderProgram,
            attribLocations: {
                vertexPosition: this.gl.getAttribLocation(shaderProgram, "aVertexPosition"),
                vertexColor: this.gl.getAttribLocation(shaderProgram, "aVertexColor")
            },
            uniformLocations: {
                projectionMatrix: this.gl.getUniformLocation(
                    shaderProgram,
                    "uProjectionMatrix"
                ),
                modelViewMatrix: this.gl.getUniformLocation(
                    shaderProgram,
                    "uModelViewMatrix"
                )
            }
        };
    }

    initVR() {
        return new Promise((resolve, reject) => {
            this.canvas.addEventListener("webglcontextlost", this.onContextLost, false);
            this.canvas.addEventListener("webglcontextrestored", this.onContextRestored, false);
            if (navigator.getVRDisplays) {
                navigator.getVRDisplays().then(
                    (displays) => {
                        if (displays.length > 0) {
                            this.vrDisplay = displays[displays.length - 1];
                            this.initGL(true);
                            if (this.vrDisplay.capabilities.canPresent) {
                                this.dispatchEvent("vrcapable");
                                window.requestAnimationFrame((now) => this.onAnimationFrame(now));
                            }
                            window.addEventListener(
                                "vrdisplaypresentchange",
                                () => this.onVRPresentChange(),
                                false
                            );
                            window.addEventListener(
                                "vrdisplayactivate",
                                () => this.requestVRPresent()
                            );
                            window.addEventListener(
                                "vrdisplaydeactivate",
                                () => this.exitVRPresent(),
                                false
                            );
                        } else {
                            this.initGL(false);
                            window.requestAnimationFrame((now) => this.onAnimationFrame(now));
                            console.log("No VRDisplays found.");
                        }
                        resolve();
                    },
                    () => {
                        this.initGL(false);
                        window.requestAnimationFrame((now) => this.onAnimationFrame(now));
                        console.log("No WebVR support.");
                        resolve();
                    }
                );
            }
        });
    }

    onAnimationFrame(now) {
        now *= 0.001; // convert to seconds
        let deltaTime = now - this.lastUpdate;
        if (this.lastUpdate == 0)
            deltaTime = 0; // first update step
        this.lastUpdate = now;

        this.update(deltaTime);
        this.render();

        if (this.vrDisplay && this.vrDisplay.isPresenting) {
            this.vrDisplay.requestAnimationFrame((now) => this.onAnimationFrame(now));
        } else {
            window.requestAnimationFrame((now) => this.onAnimationFrame(now));
        }
    }

    update(deltaTime) {
        if (this.currentScene) {
            Debug.update(deltaTime);
            this.currentScene.update(deltaTime);
            // check if any objects have been destroyed and remove them from the scene
            const tbr = this.currentScene.gameObjects.filter((go) => go.destroyed);
            tbr.forEach((go) => this.currentScene.removeObject(go));
        }
    }

    render() {
        if (this.currentScene) {
            const camera = this.currentScene.camera();
            if (!camera) {
                // no comera found. abort current render loop
                console.error("No camera in current scene");
                return;
            }

            this.canvas = document.querySelector('#glcanvas');

            this.gl.clearColor(...camera.getComponent(Camera).clearColor); // Clear to black, fully opaque
            this.gl.clearDepth(1.0); // Clear everything
            this.gl.enable(this.gl.DEPTH_TEST); // Enable depth testing
            this.gl.enable(this.gl.SAMPLE_COVERAGE); // Enable multi sampling
            this.gl.sampleCoverage(4, false);
            this.gl.depthFunc(this.gl.LEQUAL); // Near things obscure far things
            // Clear the canvas before we start drawing on it.
            this.gl.clear(this.gl.COLOR_BUFFER_BIT | this.gl.DEPTH_BUFFER_BIT);
            /* Create a perspective matrix, a special matrix that is
               used to simulate the distortion of perspective in a camera.
               Our field of view is 45 degrees, with a width/height
               ratio that matches the display size of the canvas
               and we only want to see objects between 0.1 units
               and 100 units away from the camera.
             */
            const fieldOfView = 45 * Math.PI / 180; // in radians
            const aspect = this.gl.canvas.clientWidth / this.gl.canvas.clientHeight;
            const zNear = 0.1;
            const zFar = 100.0;
            if (this.vrDisplay) {
                this.vrDisplay.depthNear = zNear;
                this.vrDisplay.depthFar = zFar;
            }
            const projectionMatrix = glMatrix.mat4.create();
            /* note: glmatrix.js always has the first argument
               as the destination to receive the result. */
            glMatrix.mat4.perspective(projectionMatrix, fieldOfView, aspect,
                zNear, zFar);
        
            const modelViewStack = new MatrixStack();
            const renderInfo = new RenderInfo(this.gl, this.programInfo, modelViewStack);
        
            // set camera position and rotation
            // create camera matrix, invert it and multiply it to out modelview matrix stack

            const invCameraMatrix = glMatrix.mat4.create();
            glMatrix.mat4.invert(invCameraMatrix, camera.transform.glMat4);
        
            // render the scene once or twice depending on if we have a vr headset or a normal browser
            if (this.vrDisplay && this.vrDisplay.isPresenting) {
                const frameData = new VRFrameData();
                this.vrDisplay.getFrameData(frameData);
                var poseMatrix = glMatrix.mat4.create();
        
                // apply the head pose to the model view stack
                getPoseMatrix(poseMatrix, frameData.pose);
                modelViewStack.save();
                // apply head pose matrix
                modelViewStack.multiply(poseMatrix);
                // move camera to camera position
                modelViewStack.multiply(invCameraMatrix);
        
                renderInfo.projectionMatrix = frameData.leftProjectionMatrix;
                // render all game objects for left eye
                this.gl.viewport(0, 0, this.canvas.width * 0.5, this.canvas.height);
                
                // render scence 
                this.currentScene.render(renderInfo);
                Debug.render(renderInfo);
        
                renderInfo.projectionMatrix = frameData.rightProjectionMatrix;
                // render all game objects for right eye
                this.gl.viewport(this.canvas.width * 0.5, 0, this.canvas.width * 0.5, this.canvas.height);
                
                // render scence 
                this.currentScene.render(renderInfo);
                Debug.render(renderInfo);
        
                // restore model view matrix
                modelViewStack.restore();
        
                this.vrDisplay.submitFrame();
            } else {
                renderInfo.projectionMatrix = projectionMatrix;
                // move camera to camera position
                modelViewStack.multiply(invCameraMatrix);
        
                // render scence 
                this.currentScene.render(renderInfo);
                Debug.render(renderInfo);
            }
        }
    }

    setCurrentScene(scene) {
        this.currentScene = scene;
    }

    setCanvasId(id) {
        this.canvasId = id;
    }

    requestVRPresent() {
        this.vrDisplay
        .requestPresent([{
            source: this.canvas
        }])
        .then(
            () => {
                // start presenting in vr
                this.dispatchEvent('presentvr');
            },
            function (err) {
                var errMsg = "requestPresent failed.";
                if (err && err.message) {
                    errMsg += "<br />" + err.message;
                }
                console.log(errMsg);
            }
        );
    }

    exitVRPresent() {
        if (!this.vrDisplay.isPresenting) {
            return;
        }
        this.vrDisplay.exitPresent().then(
            function () {
                window.requestAnimationFrame(update);
            },
            function () {
                console.log("exitPresent failed.");
            }
        );
    }

    onVRPresentChange() {
        console.log("onVRPresentChange");
        this.onResize();
        if (this.vrDisplay.isPresenting) {
            this.dispatchEvent('entervr');
        } else {
            this.dispatchEvent('exitvr');
        }
    }

    onResize() {
        console.log("onResize");
        if (this.vrDisplay && this.vrDisplay.isPresenting) {
            console.log("in VR");
            var left_eye = this.vrDisplay.getEyeParameters("left");
            var right_eye = this.vrDisplay.getEyeParameters("right");
            this.canvas.width = Math.max(left_eye.renderWidth, right_eye.renderWidth) * 2;
            this.canvas.height = Math.max(left_eye.renderHeight, right_eye.renderHeight);
            console.log(`width ${this.canvas.width}, height ${this.canvas.height}`);
        } else {
            this.canvas.width = this.canvas.offsetWidth;
            this.canvas.height = this.canvas.offsetHeight;
        }
    }
    // Draw the scene repeatedly
    onContextLost(event) {
        event.preventDefault();
        console.log("WebGL Context Lost.");
        this.gl = null;
    }

    onContextRestored(event) {
        console.log("WebGL Context Restored.");
        init(this.vrDisplay ? this.vrDisplay.capabilities.hasExternalDisplay : false);
    }
}

export const VRoomEngine = new VRoom();