import Component from "../Component.js";
import Vector3 from "../math/Vector3.js";

export default class PrimitiveRenderer extends Component{
    constructor({colors = []}) {
        super();

        this.vertices = [];
        this.indices = [];
        this.colors = colors;

        this.buffers = {
            vertices: null,
            indices: null,
            colors: null,
        }

        this.gl = null;
    }

    /**
     * Initialize the renderer. Initializes and binds the WebGL buffers for verts, indices and colors
     * 
     * @param {*} gl WebGL context
     */
    init(gl) {
        this.gl = gl;

        this.initColorsArray();

        // create buffers
        this.buffers.vertices = gl.createBuffer();
        this.buffers.indices = gl.createBuffer();
        this.buffers.colors = gl.createBuffer();

        this.updateBuffers(gl);
    }

    updateBuffers() {
        // bind vertex buffer
        this.gl.bindBuffer(this.gl.ARRAY_BUFFER, this.buffers.vertices);
        this.gl.bufferData(this.gl.ARRAY_BUFFER, new Float32Array(this.vertices), this.gl.STATIC_DRAW);
        // bind index buffer
        this.gl.bindBuffer(this.gl.ELEMENT_ARRAY_BUFFER, this.buffers.indices);
        this.gl.bufferData(this.gl.ELEMENT_ARRAY_BUFFER, new Uint16Array(this.indices), this.gl.STATIC_DRAW);
        // bind color buffer
        this.gl.bindBuffer(this.gl.ARRAY_BUFFER, this.buffers.colors);
        this.gl.bufferData(this.gl.ARRAY_BUFFER, new Float32Array(this.colors), this.gl.STATIC_DRAW);
    }

    _bindVertices(gl, programInfo) {
        /* Tell WebGL how to pull out the positions from the position
           buffer into the vertexPosition attribute
         */
        const numComponents = 3;
        const type = gl.FLOAT;
        const normalize = false;
        const stride = 0;
        const offset = 0;
        gl.bindBuffer(gl.ARRAY_BUFFER, this.buffers.vertices);
        gl.vertexAttribPointer(programInfo.attribLocations.vertexPosition,
            numComponents, type, normalize, stride, offset);
        gl.enableVertexAttribArray(
            programInfo.attribLocations.vertexPosition);
    }

    _bindColors(gl, programInfo) {
        /* Tell WebGL how to pull out the colors from the color buffer
           into the vertexColor attribute.
        */
        const numComponents = 4;
        const type = gl.FLOAT;
        const normalize = false;
        const stride = 0;
        const offset = 0;
        gl.bindBuffer(gl.ARRAY_BUFFER, this.buffers.colors);
        gl.vertexAttribPointer(programInfo.attribLocations.vertexColor,
            numComponents, type, normalize, stride, offset);
        gl.enableVertexAttribArray(
            programInfo.attribLocations.vertexColor);
    }

    _drawElements(gl) {
        // draw all tris
        gl.drawElements(gl.TRIANGLES, this.indices.length, gl.UNSIGNED_SHORT, 0);
    }

    render(renderInfo) {
        const gl = renderInfo.gl;
        const programInfo = renderInfo.programInfo;
        
        this._bindVertices(gl, programInfo);
        this._bindColors(gl, programInfo);
        // Tell WebGL which indices to use to index the vertices
        gl.bindBuffer(gl.ELEMENT_ARRAY_BUFFER, this.buffers.indices);
        // Tell WebGL to use our program when drawing
        gl.useProgram(programInfo.program);
    
        // set the projection matrix for the left eye
        gl.uniformMatrix4fv(programInfo.uniformLocations.modelViewMatrix,
            false, renderInfo.modelViewStack.getCurrentMatrix());
        gl.uniformMatrix4fv(programInfo.uniformLocations.projectionMatrix,
            false, renderInfo.projectionMatrix);
        
        this._drawElements(gl);
    }

    initColorsArray() {
        if (this.colors.length == 4) {
            // Convert the array of color values into table for all vertices
            const color = this.colors;
            this.colors = [];
            for (var i = 0; i < this.vertices.length / 3; i++) {
                this.colors.push(...color);
            }
        }
    }

    setColors(colors) {
        this.colors = colors;
        this.initColorsArray();
        this.updateBuffers();
    }

    raycast(ray) {
        const worldMat = this.transform.worldMat4;

        for(let i = 0; i < this.indices.length; i += 3) {
            const v1 = new Vector3(...this.vertices.slice(this.indices[i] * 3, this.indices[i] * 3 + 3));
            const v2 = new Vector3(...this.vertices.slice(this.indices[i + 1] * 3, this.indices[i + 1] * 3 + 3));
            const v3 = new Vector3(...this.vertices.slice(this.indices[i + 2] * 3, this.indices[i + 2] * 3 + 3));
            if (ray.intersectsTriangle([worldMat.multiply(v1), worldMat.multiply(v2), worldMat.multiply(v3)])) {
                return true;
            }
        }
        return false;
    }
}