import GameObject from "./engine/GameObject.js";
import CubeRenderer from "./engine/renderers/CubeRenderer.js";
import PlaneRenderer from "./engine/renderers/PlaneRenderer.js";
import {VRoomEngine} from "./engine/VRoom.js";
import Scene from "./engine/Scene.js";
import Camera from "./engine/Camera.js";
import CubeSpawner from "./game/CubeSpawner.js";
import LaserPointer from "./game/LaserPointer.js";
import Physics from "./engine/Physics.js";
import Vector3 from "./engine/math/Vector3.js";
import Cube from "./game/Cube.js";

function main() {
    let vrButton = null;
    VRoomEngine.setCanvasId("#glcanvas");
    
    VRoomEngine.addEventListener("vrcapable", () => {
        vrButton = addButton("Enter VR", () => { console.log("enter"); console.log(VRoomEngine); VRoomEngine.requestVRPresent() });
    });

    VRoomEngine.addEventListener("entervr", () => {
        if (VRoomEngine.vrDisplay.capabilities.hasExternalDisplay) {
            removeButton(vrButton);
            vrButton = addButton("Exit VR", VRoomEngine.exitVRPresent);
        }
    });

    VRoomEngine.addEventListener("exitvr", () => {
        if (VRoomEngine.vrDisplay.capabilities.hasExternalDisplay) {
            removeButton(vrButton);
            vrButton = addButton("Enter VR", VRoomEngine.requestVRPresent);
        }
    });

    const scene = new Scene(initScene);
    VRoomEngine.setCurrentScene(scene);

    VRoomEngine.init();
    VRoomEngine.start();
}

/*
    Init the scene.
*/
function initScene(scene) {
    const gameObjects = [];

    const gameManager = new GameObject();
    gameManager.addComponent(new CubeSpawner());
    gameObjects.push(gameManager);

    const camera = new GameObject();
    const cameraComponent = new Camera();
    cameraComponent.clearColor = [69 / 255, 180 / 255, 231 / 255, 1.0];
    camera.addComponent(cameraComponent);
    camera.transform.setPosition(0, 1.8, 5);
    camera.transform.setRotation(25, 0, 0);
    gameObjects.push(camera);
    
    const armJoint = new GameObject();
    armJoint.transform.setPosition(0.3, 1.4, 4.2);

    // const cube = new GameObject("cube");
    // cube.layer = 1;
    // cube.addComponent(new CubeRenderer());
    // cube.addComponent(new Cube());
    // cube.transform.setPosition(0, 8.6, 0);
    // gameObjects.push(cube);

    // create laser pointer
    const laserPointer = new GameObject();
    laserPointer.transform.setPosition(0, 0, 0);
    // laserPointer.transform.setRotation(35, 0, 0);
    laserPointer.addComponent(new LaserPointer());

    // create laser pointer case
    const laserPointerCase = new GameObject();
    laserPointerCase.addComponent(new CubeRenderer());
    laserPointerCase.transform.setScale(0.05, 0.05, 0.1);
    laserPointer.transform.addChild(laserPointerCase.transform);

    // create actual laser ray
    const laser = new GameObject();
    laser.addComponent(new CubeRenderer(1, [1, 0, 0, 1]));
    laser.transform.setScale(0.01, 0.01, 100);
    laser.transform.setPosition(0, 0, -50);
    laserPointer.transform.addChild(laser.transform);

    armJoint.transform.addChild(laserPointer.transform);
    gameObjects.push(armJoint);

    const ground = new GameObject();
    ground.addComponent(new PlaneRenderer(100, 100, [160 / 255, 243 / 255, 196 / 255, 1]));
    gameObjects.push(ground);

    return gameObjects;
}

function addButtonElement(message) {
    var button_elt = document.createElement("div");
    var webgl_canvas = document.getElementById('glcanvas');
    var button_container = document.getElementById("vr-button-container");
    if (!button_container) {
        button_container = document.createElement("div");
        button_container.id = "vr-button-container";
        Object.assign(button_container.style, {
            fontFamily: "sans-serif",
            position: "absolute",
            zIndex: "999",
            left: "0",
            bottom: "0",
            right: "0",
            margin: "0",
            padding: "0"
        });
        button_container.align = "right";
        webgl_canvas.parentNode.appendChild(button_container);
    }
    Object.assign(button_elt.style, {
        color: "#FFF",
        fontWeight: "bold",
        backgroundColor: "#888",
        borderRadius: "5px",
        border: "3px solid #555",
        position: "relative",
        display: "inline-block",
        margin: "0.5em",
        padding: "0.75em",
        cursor: "pointer",
    });
    button_elt.align = "center";
    button_elt.innerHTML = message;
    button_container.appendChild(button_elt);
    return button_elt;
}

function addButton(message, callback) {
    var element = addButtonElement(message);
    element.addEventListener("click", function (event) {
        callback(event);
        event.preventDefault();
    }, false);
    return {
        element: element,
    };
}

function removeButton(button) {
    if (!button) {
        return;
    }
    if (button.element.parentElement) {
        button.element.parentElement.removeChild(button.element);
    }
}

// register event listener to start the main function
window.addEventListener("load", main, false);